@@ -199,7 +199,7 @@ bool ReplicatedBackend::handle_message(
   return false;
 }
 
-void ReplicatedBackend::clear_state()
+void ReplicatedBackend::clear_recovery_state()
 {
   // clear pushing/pulling maps
   pushing.clear();
@@ -218,7 +218,7 @@ void ReplicatedBackend::on_change()
     if (i->second.on_applied)
       delete i->second.on_applied;
   }
-  clear_state();
+  clear_recovery_state();
 }
 
 void ReplicatedBackend::on_flushed()
@@ -522,7 +522,7 @@ void ReplicatedBackend::submit_transaction(
   PGTransaction *_t,
   const eversion_t &trim_to,
   const eversion_t &trim_rollback_to,
-  vector<pg_log_entry_t> &log_entries,
+  const vector<pg_log_entry_t> &log_entries,
   boost::optional<pg_hit_set_history_t> &hset_history,
   Context *on_local_applied_sync,
   Context *on_all_acked,
@@ -563,9 +563,9 @@ void ReplicatedBackend::submit_transaction(
     reqid,
     trim_to,
     trim_rollback_to,
-    t->get_temp_added().size() ? *(t->get_temp_added().begin()) : hobject_t(),
-    t->get_temp_cleared().size() ?
-      *(t->get_temp_cleared().begin()) :hobject_t(),
+    t->get_temp_added().empty() ? hobject_t() : *(t->get_temp_added().begin()),
+    t->get_temp_cleared().empty() ?
+      hobject_t() : *(t->get_temp_cleared().begin()),
     log_entries,
     hset_history,
     &op,
@@ -573,7 +573,7 @@ void ReplicatedBackend::submit_transaction(
 
   ObjectStore::Transaction local_t;
   local_t.set_use_tbl(op_t->get_use_tbl());
-  if (t->get_temp_added().size()) {
+  if (!(t->get_temp_added().empty())) {
     get_temp_coll(&local_t);
     add_temp_objs(t->get_temp_added());
   }
@@ -586,7 +586,9 @@ void ReplicatedBackend::submit_transaction(
     trim_rollback_to,
     true,
     &local_t);
-  (*op_t).append(local_t);
+
+  local_t.append(*op_t);
+  local_t.swap(*op_t);
   
   op_t->register_on_applied_sync(on_local_applied_sync);
   op_t->register_on_applied(
